/********************************************************************** 
 Freeciv - Copyright (C) 1996 - A Kjeldberg, L Gregersen, P Unold
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <player.h>
#include <city.h>
#include <game.h>
#include <unit.h>
#include <unithand.h>
#include <shared.h>
#include <cityhand.h>
#include <packets.h>
#include <map.h>
#include <mapgen.h>
#include <aitools.h>
#include <aihand.h>
#include <aiunit.h>
#include <aicity.h>


int ai_fix_unhappy(struct city *pcity);
void ai_manage_city(struct player *pplayer, struct city *);

/**************************************************************************
 cities, build order and worker allocation stuff here..
**************************************************************************/

void ai_manage_cities(struct player *pplayer)
{ 
  struct city *pcity;
  struct genlist_iterator myiter;
  genlist_iterator_init(&myiter, &pplayer->cities.list, 0);
  for(; ITERATOR_PTR(myiter); ) {
    pcity=(struct city *)ITERATOR_PTR(myiter); 
    ITERATOR_NEXT(myiter);
    ai_manage_city(pplayer, pcity);
  }

}

/**************************************************************************
... find a good (bad) tile to remove
**************************************************************************/

int worst_elvis_tile(struct city *pcity, int x, int y, int bx, int by)
{
    return (16*get_food_tile(x, y, pcity) + 
          12*get_shields_tile(x, y, pcity) + 
          8*get_trade_tile(x, y, pcity) < 
          16*get_food_tile(bx, by, pcity) + 
          12*get_shields_tile(bx, by, pcity) +
          8*get_trade_tile(bx, by, pcity) 
          );
}

void ai_do_build_unit(struct city *pcity, int unit_type)
{
  pcity->is_building_unit = 1;
  pcity->currently_building = unit_type;
}
int is_defender_unit(int unit_type) 
{
  return ((U_WARRIORS <= unit_type) && (unit_type <= U_MECH));
}

int city_get_defenders(struct city *pcity)
{
  int def=0;
  struct unit *punit;
  struct genlist_iterator myiter;
  genlist_iterator_init(&myiter, &pcity->units_supported.list, 0);
  for (; ITERATOR_PTR(myiter); ITERATOR_NEXT(myiter)) {
    punit=(struct unit*)ITERATOR_PTR(myiter);
    if (!can_build_unit(pcity, punit->type))
      continue;
    if (!is_defender_unit(punit->type))
      continue;
    if (punit->x==pcity->x && punit->y==pcity->y)
      def++;
  }
  return def;

}

int city_get_settlers(struct city *pcity)
{
  int set=0;
  struct unit *punit;
  struct genlist_iterator myiter;
  genlist_iterator_init(&myiter, &pcity->units_supported.list, 0);
  for (; ITERATOR_PTR(myiter); ITERATOR_NEXT(myiter)) {
    punit=(struct unit*)ITERATOR_PTR(myiter);
    if (unit_flag(punit->type, F_SETTLERS))
      set++;
  }
  return set;
}


int ai_in_initial_expand(struct player *pplayer, struct city *pcity)
{
  int expand_cities [3] = {3, 5, 7};
  return (pplayer->score.cities < expand_cities[get_race(pplayer)->expand]);  
}


void ai_city_build_defense(struct city *pcity)
{
  int i;
  int best= 0;
  int bestid = 0;
  for (i = U_WARRIORS; i < U_HORSEMEN ; i++) {
    if (can_build_unit(pcity, i) && 
	get_unit_type(i)->defense_strength >= best) {
      best = get_unit_type(i)->defense_strength;
      bestid = i;
    }
  }
  ai_do_build_unit(pcity, bestid);
}


void ai_city_build_settler(struct city *pcity)
{
  int i;
  for (i = 0 ; i < U_LAST ; i ++) {
    if (unit_flag(i, F_SETTLERS) && can_build_unit(pcity, i)) {
      ai_do_build_unit(pcity, i);
      break;
    }
  }
}


/************************************************************************** 
... change the build order.
**************************************************************************/

void ai_city_choose_build(struct player *pplayer, struct city *pcity)
{
  int def = city_get_defenders(pcity);
  int set = city_get_settlers(pcity);
  
  if (pcity->ai.ai_role == AICITY_NONE) {
    if (!set) {
      if (def || ai_in_initial_expand(pplayer, pcity)) {
	ai_city_build_settler(pcity); 
	return;
      }
    }
    if (!def) {
      ai_city_build_defense(pcity);
      return;
    }
    advisor_choose_build(pcity);
  }
}



/**************************************************************************
 cities, build order and worker allocation stuff here..
**************************************************************************/
void ai_manage_city(struct player *pplayer, struct city *pcity)
{
  auto_arrange_workers(pcity);
  ai_fix_unhappy(pcity);
  ai_city_choose_build(pplayer, pcity);
}

/**************************************************************************
...
**************************************************************************/

int ai_make_elvis(struct city *pcity)
{
  int x,y;
  int xp=0,yp=0;
  for (y=0;y<5;y++)
    for (x=0;x<5;x++) {
      if (((x==0 ||x==4) && (y==0 || y==4)) 
          || (x==2 && y==2))
        continue; 
      if (get_worker_city(pcity, x, y) == C_TILE_WORKER) {
        if (xp==0 && yp==0) { 
          xp=x;
          yp=y;
        } else {
          if (worst_elvis_tile(pcity, x, y, xp, yp)) {
            xp=x;
            yp=y;
          }
        }
      }
    }
  if (xp!=0 || yp!=0) {
    set_worker_city(pcity, xp, yp, C_TILE_EMPTY);
    pcity->ppl_elvis++;
    return 1;
  }
  return 0;
}


/**************************************************************************
...
**************************************************************************/

int ai_fix_unhappy(struct city *pcity)
{
  if (!is_city_unhappy(pcity))
    return 1;
  while (is_city_unhappy(pcity)) {
    if(!ai_make_elvis(pcity)) break;
    city_refresh(pcity);
  }
  return (is_city_unhappy(pcity));
}

